#!/usr/bin/perl 
#
# Copyright (c) 2005-2010  Gregor Maier <gregor@majordomus.org>
# All rights reserved. See LICENSE
# 
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#

use strict; 
use POSIX;

########### CONFIG ##################

# WARNING: You should really use an absolut path here
# the relative path is for the sake of example 

# Directory where the output files should be places
my $OUTDIR="/etc/netconfgen/output";

# Directory where copies of the output files and the 
# config should be placed
my $BAKDIR="/etc/netconfgen/backup";

# The Domainname to append to hostname for the reverse zone. 
# The last dot for the root zone is added automaticly
my $DOMAIN="example.com";

# Filename for the Forward zone. 
# NOTE: This file _only_ contains A and AAAA records. CNAMES, NS, MX 
# and other records are not handled by this file
my $DNS_v4_FILENAME="entries.hosts.v4";
my $DNS_v6_FILENAME="entries.hosts.v6";

# Filename for the v6 reverse zone 
# This file will only contain PTR records. SOA must
# be added externally
my $DNS_v6_REV_FILENAME="entries.ip6.arpa";


# The IPv4 reverse zone, that we might encounter
# i.e. multiple v4 reverse zone can be generated
# Each generated zone assumes an ORIGIN of 3 octets
#
my @IPv4_REV_ZONES;

# The length of the IPv6 Prefix for the v6 reverse zone 
# Only one IPv6 reverse zone file will be generated, with
# an expected $ORIGIN of $PREFIXLEN bit
my $PREFIXLEN=48;


# Filenames for the dhcp config files
# A seperate filename for each "dhcp-realm" that can be encountered
# must be speicifed here
my @DHCP_REALMS;

################ CONFIG ENDS ###########################

# There shouldn't be the need to change anything beyond this line. 


########### GLOBAL VARS ##############

# filehandles
my $DNSFW4; 		# forward zone handle
my $DNSFW6; 		# forward zone handle
my %DNSREV4;	# IPv4 reverse zone handles
my $DNSREV6;	# IPv6 reverse zone handle
my %DHCP_CONF;


my $errors=0;


# all flags we have:
my $dhcp;
my $NOREV;

&main;


sub main {
	my ($host, $mac, $ipv4, $ipv6type, $ipv6, $flagstr, $comment);
	my $state = "none";
	my @fhandles;
	my $ipv6addr;
	my @elements;

	while (<>) {
		chomp;
		next if /^\s*#/;   # discard comments
		next if /^\s*$/;   # discard empty lines

		if (/^\s*\[setup\]/i  && $state eq "none") {
			$state = "setup";
			next;
		}
		elsif (/^\s*\[setup\]/i) {
			die "FATAL: There was a [setup] statement where none should be (while in state $state)\n";
		}
		if (/^\s*\[hosts\]/i  && $state eq "setup") {
			$state = "hosts";
			@fhandles = finish_setup();
			next;
		}
		elsif (/^\s*\[hosts\]/i) {
			die "FATAL: $_ There was a [hosts] statement where none should be (while in state $state)\n";
		}

		# Handle a line in setup state
		if ($state eq "setup") {
			handle_setup_line($_);
		}

		if ($state eq "hosts") {
			# $_ =~ s/(.*)#.*/\1/;  # Remove commets

			@elements = split /\|/, $_, 7; # Split line into tokens
			# Remoe leading and trailing whitespaces for each token
			foreach (@elements) {

				 s/^\s*(\S*)\s*$/\1/;
			}
			($host, $mac, $ipv4, $ipv6type, $ipv6, $flagstr, $comment) = @elements;

			# Split the flags 
			my @flagsarr = undef;
			@flagsarr = split /,/, $flagstr;
			$dhcp = "";
			$NOREV=0;
			foreach my $curflag (@flagsarr) {
				# Remove leading and trailing whitespaces
				$curflag =~ s/^\s*(\S*)\s*$/\1/;

				
				# If the flag is all lowercase, it's the dhcp-realm
				if ($curflag =~ /^[a-z][a-z0-9]*$/) {
					$dhcp=$curflag;
				}
				elsif ($curflag eq "NOREV") {
					$NOREV="1";
				}
				else { # unknown Flag
					print STDERR "ERROR: flag \"$curflag\" for host $host is unknown\n";
					$errors++;
				}
			} # end split flags

			$ipv6addr = get_ipv6_addr($ipv6type, $ipv6, $mac, $host);
			print_dhcp($host, $mac, $dhcp);
			print_dns_v4_forward($host,$ipv4);
			print_dns_v6_forward($host,$ipv6addr);
			print_dns_v4_rev($host, $ipv4);
			print_dns_v6_rev($host, $ipv6addr);
		}

	}
	foreach (@fhandles) {
		close $_;
	}
} # end main


sub handle_setup_line {
	my $line = shift;

	my ($name, $value) = split(/=/, $line);
	# remove leading and trailing whitespace
	$name =~ s/^\s*(\S*)\s*$/\1/;
	$value =~ s/^\s*(\S*)\s*$/\1/;

	if ($name eq "outputdir") {
		$OUTDIR = $value;
	} elsif ($name eq "backupdir") {
		$BAKDIR = $value;
	} elsif ($name eq "domain") {
		$DOMAIN = $value;
	} elsif ($name eq "dns_v4_filename") {
		$DNS_v4_FILENAME = $value;
	} elsif ($name eq "dns_v6_filename") {
		$DNS_v6_FILENAME = $value;
	} elsif ($name eq "dns_v6_rev_filename") {
		$DNS_v6_REV_FILENAME = $value;
	} elsif ($name eq "ipv6_prefixlen") {
		$PREFIXLEN = $value;
	} elsif ($name eq "dns_v4_rev_zones") {  # multiple
		push @IPv4_REV_ZONES, $value;
	} elsif ($name eq "dhcp_realms") { # multiple
		push @DHCP_REALMS, $value;
	} else {
		die "FATAL: Unknown options in [setup] state \"$line\"\n";
	}
}


sub do_backup {
	if (! -d $BAKDIR) {
		die "FATAL: Backup directory $BAKDIR does not exist";
	}

	my @ts=localtime(time());	
	my $datestr = strftime("%Y%m%d--%H%M%S", @ts);

	system("mkdir $BAKDIR/$datestr");
	system("mv $OUTDIR/* $BAKDIR/$datestr/");
	#system("cp $INPUT $BAKDIR/$datestr/");
}


#
# The setup sections in the input file ends
# i.e. the hosts section will start now. we can now open all
# the required output files
sub finish_setup() {
	my $fhandles;

	&do_backup;

	if (!-d $OUTDIR) { 
		die "FATAL: Basedirectory \"$OUTDIR\"  does not exist!\n";
	}

	my @fhandles;
	open $DNSFW4, "> $OUTDIR/$DNS_v4_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v4_FILENAME for writing\n";
	push @fhandles, $DNSFW4;

	open $DNSFW6, "> $OUTDIR/$DNS_v6_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v6_FILENAME for writing\n";
	push @fhandles, $DNSFW6;

	open $DNSREV6, "> $OUTDIR/$DNS_v6_REV_FILENAME" or 
			die "FATAL: Could not open $OUTDIR/$DNS_v6_REV_FILENAME for writing\n";
	push @fhandles, $DNSREV6;

	foreach (@IPv4_REV_ZONES) {
		my ($zone,$fn) = split(/:/, $_);
		open $DNSREV4{$zone}, "> $OUTDIR/$fn" or
			die "FATAL: Could not open $OUTDIR/$fn for writing\n";
		push @fhandles, $DNSREV4{$zone};
	}

	# @fhandles now contains all DNS/Bind related filehandles we will ever
	# have open. So print a nice message to each file reminding the user
	# not to mess with the automaticly gernated file!
	foreach (@fhandles) {
		print $_ "; AUTOMATICLY GENERATED FILE -- DO NOT EDIT\n";
		print $_ "; I REALLY MEAN DO NOT EDIT!!\n";
	}

	foreach (@DHCP_REALMS) {
		my ($realm,$fn) = split(/:/, $_);
		open $DHCP_CONF{$realm},  "> $OUTDIR/$fn" or
			die "FATAL: Could not open $OUTDIR/$fn  for writing\n";
		my $tmpfh = $DHCP_CONF{$realm};
		print $tmpfh "# AUTOMATICLY GENERATED FILE -- DO NOT EDIT\n";
		print $tmpfh "# I REALLY MEAN DO NOT EDIT!!\n";
		push @fhandles, $DHCP_CONF{$realm};
	}
	return @fhandles;
}


sub get_ipv6_addr {
	my ($ipv6type, $ipv6, $mac, $host) = @_;

	my $ipv6addr="";
	if ($ipv6type eq "fix") {
		$ipv6addr = $ipv6;
	}
	elsif ($ipv6type eq "auto") {
		my $tmpaddr = `ipv6calc --in mac $mac --action geneui64 --out eui64`;
		chomp $tmpaddr;
		if ($tmpaddr =~ /[0-9a-fA-F:]+/) {
			$ipv6addr = "$ipv6:$tmpaddr";
		}
		else {
			print STDERR "ERROR: ipv6calc generated stupid results for host $host\n";
			$errors++;
		}
	}
	elsif ($ipv6type eq "none") {
		; # do nothing
	}
	else {
		print STDERR "ERROR: ipv6 type \"$ipv6type\" for host $host is unknown\n";
		$errors++;
	}
}

# DNS FORWARD ZONE
sub print_dns_v4_forward ($$) {
	my ($host, $ipv4) = @_;

	print $DNSFW4 "$host\tIN A\t$ipv4\n";
}

sub print_dns_v6_forward ($$) {
	my ($host, $ipv6addr) = @_;

	if ($ipv6addr ne "") {
		print $DNSFW6 "$host\tIN AAAA\t$ipv6addr\n";
	}
}

sub print_dns_v4_rev ($$){
	my ($host, $ipv4) = @_;
	# DNS IPv4 REVERSE ZONE(S)
	my @octets;
	my $revzone;
	if (!$NOREV) { 
		@octets = split /\./, $ipv4;
		$revzone = "$octets[2].$octets[1].$octets[0].in-addr.arpa";
		my $fh=$DNSREV4{$revzone};
		print $fh "$octets[3]\tIN PTR\t$host.$DOMAIN.\n";
	}
	else {
		print STDERR "NOTE: no reverse IPv4 for $host\n";
	}
}

sub print_dns_v6_rev ($$) { 
	my ($host, $ipv6addr) = @_;


	# DNS IPv6 REVERSE ZONE
	my $revaddr;
	if (!$NOREV && ($ipv6addr ne "")) { 
		$revaddr = `ipv6calc --in ipv6addr $ipv6addr --out revnibbles.arpa`;
		chomp $revaddr;
		if ($revaddr =~ /^([a-fA-F0-9]\.){32}ip6.arpa.$/ ) {
				# each 4 Bits of the address is 2 characters wide (nibble + dot)
				# therefor: the number of nibbles we need is  (128-prefixlen)/4
				# then multiply with two to account for the dot 
				# then subtract 1 to get rid of the trailing dot.
				$revaddr = substr($revaddr, 0, (128-$PREFIXLEN)/2-1); 
				print $DNSREV6 "$revaddr\tIN PTR\t$host.$DOMAIN.\n";
		}
		else { # ipv6calc returned bogus data
			print STDERR "ERROR: ipv6calc generated stupid results for host $host (reverse entry)\n";
			$errors++;
		}
		
	}
}


sub print_dhcp ($$$) {
	my ($host, $mac, $dhcp) = @_;
	
	if ($dhcp ne "") {
		if ($DHCP_CONF{$dhcp} == undef) {
			print STDERR "ERROR: no dhcp output file for dhcp realm \"$dhcp\" defined. Host was $host\n";
			$errors++;
			return;
		}
		my $tmpfh = $DHCP_CONF{$dhcp};
		print $tmpfh "host $host {\n";
		print $tmpfh "    hardware ethernet $mac;\n";
		print $tmpfh "    fixed-address $host.$DOMAIN;\n";
		print $tmpfh "}\n";
	}
}
